﻿##########################################
# package Date
#   ver1.0.0 : 2021/7/15
#   coded by ceder
##########################################
# date expressions
#   yy,yy/mm,yy/mm/dd
#   yy:4digits,(m|t|s|h|r)digits
#   mm,dd:1 or 2 digits
#   yy should be larger than 1873(m6)

package Date;
use v5.14;
use strict;
use warnings;
use utf8;

use Time::Local 'timelocal';

##########################################
##    new
##    put
##    get
##    get_ymd
##    printDate
##########################################

#  new('date expression')
#  new(Date-type instance)

sub new {
 my $class=shift;
 my $inp = shift;

 my $date; 
 my ($year,$month,$day);

 if( ! ref($inp)) {
  if( $inp =~ /\A(([RrHhSsTtMm])?[0-9]+)(\/([0-9]{1,2})(\/([0-9]{1,2}))?)?\z/ ) {

    if(defined $4 && ($4<1 || $4>12) ) {
        return undef;
      };
    if(defined $6 && ($6 > (31,28,31,30,31,30,31,31,30,31,30,31)[$4-1]) ) {
      if(($4 == 2) && ($6 == 29)) {
        my $seireki=$1;
         if( $seireki=~m/\A(?<nengou>[RrHhSsTtMm])(?<nen>[0-9]+)\z/ ) {  
           my $nen1=$+{nen};
           given( $+{nengou} ) {
              when ( /[Rr]/ ) { $seireki = 2018 + $nen1; }
              when ( /[Hh]/ ) { $seireki = 1988 + $nen1; }
              when ( /[Ss]/ ) { $seireki = 1925 + $nen1 ; }
              when ( /[Tt]/ ) { $seireki = 1911 + $nen1 ; }
              when ( /[Mm]/ ) { $seireki = 1867 + $nen1 ; }
            default {
               warn "new:abnormal error - $_\n";
               exit;
             };
            };
           };
         if((($seireki%4 != 0) || ($seireki%100 == 0 )) &&  ($seireki%400 != 0)) {
             return undef;
             };
        } else {
          return undef;
       };
     } ;
 
    $date={
     'year'=>$1,
     'month'=>$4,
     'day'=>$6
       };

     if(defined $2) {
       $date->{'ytype'}=2;
      } else {
       $date->{'ytype'}=1;
      };

    } else  {
      return undef;
    };
 
   } elsif(ref($inp) eq 'Date') {
    $date={ %{$inp}};
   } else {
    return undef;
   };

   bless $date,'Date';

 }

# put('key',value)
sub put {
  my $self=shift;
  my $key=shift;
  my $value=shift;
  $self->{$key} = $value;
   };

# get('key')
sub get {
  my $self=shift;
  my $key=shift;
  $self->{$key} ;
   }  ;

sub get_keys {
  my $self = shift ;
  keys (%$self);
  };


# get(Date-type instance)
sub get_ymd {
  my $self=shift ;
  ($self->{'year'},$self->{'month'},$self->{'day'});
}

#printDate

sub printDate{
  my $self=shift;

  my ($year,$month,$day);

  $year=$self->{'year'};
  $month=$self->{'month'};
  $day=$self->{'day'};

  if($day) {
    return sprintf("%s/%d/%d",$year,$month,$day);
   } elsif($month) {
     return  sprintf("%s/%d",$year,$month);
   } elsif($year) {
     return $year;
   } else {
     return "";
   };
 }

 sub copy{
   my $self=shift;
   my $src=shift;

   if(ref($src) eq 'Date') {
    foreach my $key (keys %$src) {
      $self->{$key}=$src->{$key};
     };
    };
   }
 


###################################
## convertYear ($type));
##  $type=0(default)
##    change Date instance   
##    in which 'year':yyyy->Xyy or Xyy->yyyy
##  $type=1
##    change Date instance such as 'year' in yyyy.
##  $type=2 
##    change Date instance such as 'year' in Xyy)
####################################

sub convertYear {
my $self=shift;
my $type = shift;
# $type = $type//0;
   $type //=0;
my $year;

#print "type=",$type,"\n";

$_=$self->{'year'};

if( m/\A(?<nengou>[RrHhSsTtMm])(?<nen>[0-9]+)\z/ ) {  
 if($type == 2) {
    $year=$_;
  } else {
   my $nen1=$+{nen};
   given( $+{nengou} ) {
     when ( /[Rr]/ ) { $year = 2018 + $nen1; }
     when ( /[Hh]/ ) { $year = 1988 + $nen1; }
     when ( /[Ss]/ ) { $year = 1925 + $nen1 ; }
     when ( /[Tt]/ ) { $year = 1911 + $nen1 ; }
     when ( /[Mm]/ ) { $year = 1867 + $nen1 ; }
     default {
      warn "convertYear:illegal expression - $_\n";
      $year = undef;
      }
    };
   };
  } elsif ( m/\A(?<nen>[0-9]{4})\z/ ) {
   if($type == 1) {
      $year = $_;
   } else {    
    if( $+{nen} > 2018 ) {
      $year = "R".sprintf("%d",$+{nen}-2018); 
    } elsif ( $+{nen} > 1988 ) {
       $year = "H".sprintf("%d",$+{nen}-1988);
      } elsif ( $+{nen} > 1925 ) {
        $year = "S".sprintf("%d",$+{nen}-1925); 
       } elsif ( $+{nen} > 1911 ) {
          $year = "T".sprintf("%d",$+{nen}-1911); 
        } elsif ( $+{nen} > 1867 ) { 
           $year =  "M".sprintf("%d",$+{nen}-1867);
          } else {
              warn "convertYear:too small A.D.";
              $year = undef;
          };
        };
     } else {
    warn "convertYear:illegal format";
    $year = undef;
   };
#  print "year=",$year,"\n";
  $self->{'year'}=$year;
  $self->{'ytype'}=$type;

  }

###################################
## afterYears ($years,$type));
##  $type=0(default)
##    excluding self year 
##  $type=1
##    including self year
####################################

sub afterYears {
my $self=shift;
my $years =shift;
my $type = shift;
#$years = $years//1;
#$type = $type//0;
$years //= 1;
$type //= 0;

   my $ytype=$self->{'ytype'};
   $self->convertYear(1);

   my $year=$self->{'year'};
   my $month=$self->{'month'};
   my $day=$self->{'day'};

   $years-- if $type;   

   $year+=$years;

   if(defined($month) && defined($day) ) {
    if($month == 2 && $day == 29 ) {
     unless((($year%4 == 0) && ($year%100 != 0 )) || $year%400 == 0) {
      $day=28;
     };
    }; 
   };

   $self->{'year'}=$year;
   $self->{'day'}=$day;

  $self->convertYear($ytype);
   
 } 

###################################
## beforeYears ($days,$type));
##  $type=0(default)
##    excluding self date 
##  $type=1
##    including self date
####################################

sub beforeYears {
my $self=shift;
my $years =shift;
my $type = shift;
#$years = $years//1;
#$type = $type//0;
$years //= 1;
$type //= 0;

   my $ytype=$self->{'ytype'};
   $self->convertYear(1);

   my $year=$self->{'year'};
   my $month=$self->{'month'};
   my $day=$self->{'day'};

   $years-- if $type;   
   $year-=$years;

  if(defined($month) && defined($day) ) {
   if($month == 2 && $day == 29 ) {
    unless(($year%4 == 0 && ($year%100 != 0 )) || $year%400 == 0) {
      $day=28;
     };
    }; 
   };

   $self->{'year'}=$year;
   $self->{'day'}=$day;

  $self->convertYear($ytype);
   
 } 

###################################
## afterDays ($days,$type));
##  $type=0(default)
##    excluding self date 
##  $type=1
##    including self date
####################################

sub afterDays {
my $self=shift;
my $days =shift;
my $type = shift;
#$days = $days//1;
#$type = $type//0;
$days //= 1;
$type //= 0;
my ($sec,$min,$hour,$year,$month,$day);

 if(defined ($day=$self->{'day'})) {
   my $ytype=$self->{'ytype'};
   $self->convertYear(1);

   $year=$self->{'year'};
   $month=$self->{'month'}-1;

   my $unixtime=timelocal(0,0,0,$day,$month,$year);
   
   $days-- if $type;
#   if($type) { 
#     $days--;
#     };
   my $incre=$days*24*3600;

   ($sec,$min,$hour,$day,$month,$year)=localtime($unixtime+$incre);

   $self->{'year'}=$year+1900;
   $self->{'month'}=$month+1;
   $self->{'day'}=$day;

  $self->convertYear($ytype);
   
  } else {
  warn "afterDays:'day' is undef. No operation.\n";
  };
 } 


###################################
## beforeDays ($days,$type));
##  $type=0(default)
##    excluding self date 
##  $type=1
##    including self date
####################################

sub beforeDays {
my $self=shift;
my $days =shift;
my $type = shift;
#$days = $days//1;
#$type = $type//0;
$days //= 1;
$type //= 0;

my ($sec,$min,$hour,$year,$month,$day);
 
 if(defined ($day=$self->{'day'})) {
   my $ytype=$self->{'ytype'};
   $self->convertYear(1);

   $year=$self->{'year'};
   $month=$self->{'month'}-1;

   my $unixtime=timelocal(0,0,0,$day,$month,$year);

 $days-- if $type;
#   if($type) { 
#     $days--;
#     };

   my $decre=$days*24*3600;

   ($sec,$min,$hour,$day,$month,$year)=localtime($unixtime-$decre);

   $self->{'year'}=$year+1900;
   $self->{'month'}=$month+1;
   $self->{'day'}=$day;
   $self->convertYear($ytype); 
  } else {
  warn "beforeDays:'day' is undef. No operation.\n";
  };
 } 

############################################################
#    Below only treating year/month case.
#        value for the key 'day' should be undef.
############################################################

my $BASEYEAR=1800;

# sub getAbsMon($year,$month) #
sub getAbsMon {
  my $year =shift;
  my $month=shift;

  ($year-$BASEYEAR) * 12 + $month;
 }

# sub getYearMon($abs_month) #
sub getYearMon {
  my $abs_month=shift;
  my ($year,$month);
  $year = int($abs_month/12)+$BASEYEAR;
  $month = $abs_month%12;
  $year = $month ? $year : $year-1;
  $month =$month ? $month : 12;

  return ($year,$month); 

}

###################################
## afterMonths ($months,$type));
##  $type=0(default)
##    excluding self month 
##  $type=1
##    including self month
####################################

sub afterMonths {
my $self=shift;
my $months =shift;
my $type = shift;
#$months = $months//1;
#$type = $type//0;

$months //=1;
$type //= 0;

 if(defined($self->{'day'}) ) {
   warn "afterMonths:'day' is specified. The result may be incorrect.\n";
   };

 my $ytype=$self->{'ytype'};
 $self->convertYear(1);
 
 my ($year,$month)=($self->{'year'},$self->{'month'});

 my $absmon=getAbsMon($year,$month);

 $months-- if $type ;
 #if($type) {
 #  $months--;
 # };

 $absmon+=$months;
  
 ($self->{'year'},$self->{'month'})=getYearMon($absmon);

 $self->convertYear($ytype);

}

###################################
## beforeMonths ($months,$type));
##  $type=0(default)
##    excluding self month 
##  $type=1
##    including self month
####################################

sub beforeMonths {
my $self=shift;
my $months =shift;
my $type = shift;
#$months = $months//1;
#$type = $type//0;
$months //=1;
$type //= 0;

 if(defined($self->{'day'}) ) {
   warn "beforeMonths:'day' is specified. The result may be incorrect.\n";
   };

 my $ytype=$self->{'ytype'};
 $self->convertYear(1);
 
 my ($year,$month)=($self->{'year'},$self->{'month'});

 my $absmon=getAbsMon($year,$month);

 $months-- if $type ;

#if($type) {
#   $months--;
#  };

 $absmon-=$months;
  
 ($self->{'year'},$self->{'month'})=getYearMon($absmon);

 $self->convertYear($ytype);

}
 1;
 