##########################################
# package Dateutil
#   ver2.0.0 : 2021/7/15
#   coded by ceder
##########################################
# utility subroutines for Date-type
#

package Dateutil;

use v5.14;
use warnings;
use strict;
use utf8;

use Date;
use Exporter qw(import);
our @EXPORT = qw( countMonth countYear countDay compareDate getAge);

use Time::Local 'timelocal';

###################################
## countYear ($fromdate,$todate);
##    $fromdate :Date instance
##    $todate :Date instance
##   return:number of Month between $fromdate to $todate 
##      if $fromdate==$todate then returned value is 1  
##
##   'month' and 'day' should be undef
####################################


sub countYear {
 my $fromdate=shift;
 my $todate=shift;

 my $from=Date->new($fromdate);
 my $to=Date->new($todate);


 $from->convertYear(1);
 $to->convertYear(1);

 my ($fyear,$fmon,$fday)=$from->get_ymd;
 my ($tyear,$tmon,$tday)=$to->get_ymd;

 if(defined($fmon) || defined($tmon) || defined($fday) || defined ($tday) ) {
   warn "countYear:'month' and/or 'day' is specified. The result may be incorrect.\n";
   };
 
  if ($fyear > $tyear ) {
   warn "countYear: from_value larger than to_value ";
   return undef;
  } else {
   return $tyear-$fyear+1;
 };

}



###################################
## countMonth ($fromdate,$todate);
##    $fromdate :Date instance
##    $todate :Date instance
##   return:number of Month between $fromdate to $todate 
##      if $fromdate==$todate then returned value is 1  
##
##   'day' should be undef
####################################


sub countMonth {
 my $fromdate=shift;
 my $todate=shift;

 my $from=Date->new($fromdate);
 my $to=Date->new($todate);


 $from->convertYear(1);
 $to->convertYear(1);

 my ($fyear,$fmon,$fday)=$from->get_ymd;
 my ($tyear,$tmon,$tday)=$to->get_ymd;

 if(defined($fday) || defined ($tday) ) {
   warn "countMonth:'day' is specified. The result may be incorrect.\n";
   };
 
 if(!defined($fmon) || !defined($tmon) ) {
      warn "countMonth:'month' is not specified. assumed 1.\n";
      $fmon//=1;
      $tmon//=1;
   };

 my $fmonth=Date::getAbsMon($fyear,$fmon);
 my $tmonth=Date::getAbsMon($tyear,$tmon);

 if ($fmonth > $tmonth ) {
   warn "countMonth: from_value larger than to_value ";
   return undef;
  } else {
   return $tmonth-$fmonth+1;
 };

}

#

###################################
## countDay ($fromdate,$todate);
##    $fromdate :Date instance
##    $todate :Date instance
##   return:number of Month between $fromdate to $todate 
##      if $fromdate==$todate then returned value is 1  
##  
####################################

sub countDay {
 my $fromdate=shift;
 my $todate=shift;

 my $from=Date->new($fromdate);
 my $to=Date->new($todate);


 $from->convertYear(1);
 $to->convertYear(1);

 my ($fyear,$fmon,$fday)=$from->get_ymd;
 my ($tyear,$tmon,$tday)=$to->get_ymd;

 if(!defined($fday) || !defined ($tday) ) {
   warn "countDay:'day' is not specified. assumed 1.\n";
     $fday//=1;
     $tday//=1;
   };
 
 if(!defined($fmon) || !defined($tmon) ) {
      warn "countDay:'month' is not specified. assumed 1.\n";
      $fmon//=1;
      $tmon//=1;
   };

 my $funixtime=timelocal(0,0,0,$fday,$fmon-1,$fyear);
 my $tunixtime=timelocal(0,0,0,$tday,$tmon-1,$tyear);

 if ($funixtime > $tunixtime) {
   warn "countDay: from_value larger than to_value ";
   return undef;
  } else {
   return ($tunixtime-$funixtime)/(24*3600)+1;
 };

}


###################################
## compareDate ($adate,$bdate,$mode);
##    $adate :Date instance
##    $bdate :Date instance
##     if 'day' not exist,it is assumed as 1
##     if 'month' not exist,it is assumed as 1
##    $mode : 0 :ignore 'day' even if it exists
##          : 1 or omitted : compare 'day' if exists
##    return
##      0 if $adate == $bdate
##      1 if $adate > $bdate
##     -1 if $adate < $bdate
##
####################################


sub compareDate {
 my $adate = shift;
 my $bdate = shift;
 my $mode =shift;
 $mode = $mode//1;

 my $a=new Date($adate);
 my $b=new Date($bdate);

 $a->convertYear(1);
 $b->convertYear(1);

 my ($ayear,$amon,$aday)=$a->get_ymd;
 my ($byear,$bmon,$bday)=$a->get_ymd;
 $amon//=1;
 $bmon//=1;
 $aday//=1;
 $bday//=1;

  my $aval= Date::getAbsMon($ayear,$amon);
  my $bval= Date::getAbsMon($byear,$bmon);
  
   if ($aval > $bval ) {
      return 1;
    } elsif ($aval <$bval ) {
      return -1;
    } else { 
     if($mode) {    
       if ($aday > $bday ) {
        return 1;
       } elsif($aday < $bday ) {
         return -1;
       } else {
        return 0;
       };
      } else {
       return 0;
      };
    }; 
 }  


###################################
## getAge ($birthday,$date(,$ret))
##   $birthday:Date instance
##   $date :Date instance
##   $ret:int.(default=0)    
##    
##  return age when $ret=0 
##  return (age,month) otherwise
##     age is 'legal age'  
##
##  When 'day'  of the date is undef, it is assumed that  
##   the month anniversary (the day before $bithday) has passed.
##   However when $birthday is the first day ,it is provided that the day 
##   is not the last day of the month. 
#################################### 
sub getAge {
  my $bd=shift;
  my $date=shift;
  my $ret=shift;
  $ret=$ret//0; 

  my $bd1=Date->new($bd);
  my $date1=Date->new($date);

  $bd1->convertYear(1);
  $date1->convertYear(1);

  my ($byear,$bmon,$bday)=$bd1->get_ymd;
  my ($year,$mon,$day)=$date1->get_ymd;

  unless(defined($byear) && defined($bmon) && defined($bday)) {
   warn "getAge:illegal birthday";
   return undef;
   };

  unless(defined($year) && defined($mon)) {
   warn "getAge:illegal date";
   return undef;   
   };

  my $bmonth=Date::getAbsMon($byear,$bmon);
  my $month=Date::getAbsMon($year,$mon);

  if($bmonth>$month) {
   warn "getAge:birthday later than date";
   return undef;
   };
 
  my $months=$month-$bmonth;

# $day`Ăꍇ
#   aP̏ꍇ
#     ̏ꍇ(̏ꍇj $months+1
#     ȊȌꍇ́@$months
#   aPȊȌꍇ
#     ȃOȌ̏ꍇ@$months@     @@@
#     ȃOȌꍇ
#       ̏ꍇ@$months y폜z
#       ȊȌꍇ@$months-1
# $day`ĂȂꍇ 
# @@aP̏ꍇ $months+1@y폜z
#     ȊȌꍇ@$months

  

  if(defined($day)) {
    if($bday == 1) {
      my $nextday=Date->new($date1);
      $nextday->afterDays;
      if($nextday->get('month') != $mon) {
        $months++; };
      } else {
       if( $day < $bday-1 ) {
#         my $nextday=Date->new($date1);
#         $nextday->afterDays;        
#         if($nextday->get('month') == $mon) {
         $months--; 
#         };
       };
      };
    } else {
#     $month++ if $bday == 1;
    };
  
  my $y=int($months/12);
  my $m=$months%12;

  if( $ret==0) {
       return $y;
   } else {
       return ($y,$m);
   };

 }

1;


      
         
 

 